/**
 * Copyright (c) 2025 Dominic Masters
 * 
 * This software is released under the MIT License.
 * https://opensource.org/licenses/MIT
 */

#include "entity.h"
#include "assert/assert.h"
#include "input.h"
#include "display/render.h"
#include "world/world.h"
#include "physics/physics.h"

#include "ui/uitextbox.h"

inventory_t PLAYER_INVENTORY;

void playerInit() {
  entity_t *ent = &ENTITIES[0];

  entity_t playerEntityData = {
    .id = PLAYER_ENTITY_ID,
    .type = ENTITY_TYPE_PLAYER,
    .x = WORLD_PLAYER_SPAWN_X,
    .y = WORLD_PLAYER_SPAWN_Y,
  };

  entityLoad(ent, &playerEntityData);
  inventoryInit(&PLAYER_INVENTORY, INVENTORY_SIZE_MAX);
}

void playerEntityLoad(entity_t *entity, const entity_t *source) {
  assertNotNull(entity, "Entity pointer cannot be NULL");
  assertNotNull(source, "Source entity pointer cannot be NULL");
  assertTrue(entity->type == ENTITY_TYPE_PLAYER, "Entity type must be PLAYER");
  assertTrue(source->type == entity->type, "Source/Entity type mismatch");
}

void playerEntityUpdate(entity_t *entity) {
  assertNotNull(entity, "Entity pointer cannot be NULL");
  assertTrue(entity->type == ENTITY_TYPE_PLAYER, "Entity type must be PLAYER");

  if(UI_TEXTBOX.visible) {
    entity->vx = entity->vy = 0;
    return;
  }

  if(inputIsDown(INPUT_BIND_UP)) {
    if(inputIsDown(INPUT_BIND_LEFT)) {
      entity->vx = -PLAYER_MOVE_SPEED_XY;
      entity->vy = -PLAYER_MOVE_SPEED_XY;

      if(entity->dir != ENTITY_DIR_NORTH && entity->dir != ENTITY_DIR_WEST) {
        entity->dir = ENTITY_DIR_NORTH;
      }
    } else if(inputIsDown(INPUT_BIND_RIGHT)) {
      entity->vx = PLAYER_MOVE_SPEED_XY;
      entity->vy = -PLAYER_MOVE_SPEED_XY;

      if(entity->dir != ENTITY_DIR_NORTH && entity->dir != ENTITY_DIR_EAST) {
        entity->dir = ENTITY_DIR_NORTH;
      }
    } else {
      entity->vy = -PLAYER_MOVE_SPEED;
      entity->vx = 0;
      entity->dir = ENTITY_DIR_NORTH;
    }
  } else if(inputIsDown(INPUT_BIND_DOWN)) {
    if(inputIsDown(INPUT_BIND_LEFT)) {
      entity->vx = -PLAYER_MOVE_SPEED_XY;
      entity->vy = PLAYER_MOVE_SPEED_XY;

      if(entity->dir != ENTITY_DIR_SOUTH && entity->dir != ENTITY_DIR_WEST) {
        entity->dir = ENTITY_DIR_SOUTH;
      }
    } else if(inputIsDown(INPUT_BIND_RIGHT)) {
      entity->vx = PLAYER_MOVE_SPEED_XY;
      entity->vy = PLAYER_MOVE_SPEED_XY;

      if(entity->dir != ENTITY_DIR_SOUTH && entity->dir != ENTITY_DIR_EAST) {
        entity->dir = ENTITY_DIR_SOUTH;
      }
    } else {
      entity->vy = PLAYER_MOVE_SPEED;
      entity->vx = 0;
      entity->dir = ENTITY_DIR_SOUTH;
    }
  } else if(inputIsDown(INPUT_BIND_LEFT)) {
    entity->vx = -PLAYER_MOVE_SPEED;
    entity->vy = 0;
    entity->dir = ENTITY_DIR_WEST;
  } else if(inputIsDown(INPUT_BIND_RIGHT)) {
    entity->vx = PLAYER_MOVE_SPEED;
    entity->vy = 0;
    entity->dir = ENTITY_DIR_EAST;
  } else {
    entity->vx = 0;
    entity->vy = 0;
  }

  // Interact
  if(inputPressed(INPUT_BIND_ACTION)) {
    entity_t *other = ENTITIES;
    do {
      if(other == entity || other->type == ENTITY_TYPE_NULL) {
        other++;
        continue;
      }

      // Is the other entity interactable?
      if(ENTITY_CALLBACKS[other->type].interact == NULL) {
        other++;
        continue;
      }

      fixed248_t distanceX = fx248Subfx248(other->x, entity->x);
      fixed248_t distanceY = fx248Subfx248(other->y, entity->y);
      fixed248_t distance = fx248Sqrt(fx248Addfx248(
        fx248Mulfx248(distanceX, distanceX),
        fx248Mulfx248(distanceY, distanceY)
      ));

      if(distance > PLAYER_INTERACT_RANGE) {
        other++;
        continue;
      }

      // Get angle
      fixed248_t angle = fx248Atan2(distanceY, distanceX);
      while(angle < 0) angle += FX248_PI;
      fixed248_t selfAngle = entityDirToAngle(entity->dir);
      while(selfAngle < 0) selfAngle += FX248_PI;
      
      // Check if angle is within range
      fixed248_t angleDiff = fx248Subfx248(angle, selfAngle);
      if(angleDiff > FX248_HALF_PI) angleDiff -= FX248_PI;
      if(angleDiff < -FX248_HALF_PI) angleDiff += FX248_PI;
      if(fx248Abs(angleDiff) > PLAYER_INTERACT_ANGLE) {
        other++;
        continue;
      }

      ENTITY_CALLBACKS[other->type].interact(entity, other);
      entity->vx = 0;
      entity->vy = 0;
      other++;
    } while(other != ENTITIES + ENTITY_COUNT_MAX);
  }
}

fixed248_t entityDirToAngle(const entitydir_t dir) {
  switch(dir) {
    case ENTITY_DIR_NORTH: return FX248_HALF_PI;
    case ENTITY_DIR_SOUTH: return -FX248_HALF_PI;
    case ENTITY_DIR_EAST: return 0;
    case ENTITY_DIR_WEST: return FX248_PI;
    default: return 0; // Should never happen
  }
}